# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import logging
import os
import pickle  # nosec  # pylint: disable=import_pickle
import random
from hashlib import md5
from collections import defaultdict

import numpy as np

from .... import opcodes
from .... import tensor as mt
from ....config import options
from ....core import ENTITY_TYPE, OutputType, recursive_tile
from ....core.context import get_context
from ....core.operand import OperandStage
from ....lib.filesystem import get_fs, FileSystem
from ....serialization.serializables import (
    KeyField,
    StringField,
    Int32Field,
    Int64Field,
    DictField,
    AnyField,
    BytesField,
    BoolField,
)
from ....tensor.core import TensorOrder
from ....utils import has_unknown_shape, Timer, ceildiv
from ...operands import LearnOperand, LearnOperandMixin
from ..core import proxima, validate_tensor, get_proxima_type

logger = logging.getLogger(__name__)


class ProximaSearcher(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.PROXIMA_SIMPLE_SEARCHER
    _tensor = KeyField("tensor")
    _distance_metric = StringField("distance_metric")
    _dimension = Int32Field("dimension")
    _row_number = Int64Field("row_number")
    _topk = Int32Field("topk")
    _threads = Int32Field("threads")
    _index = AnyField("index")
    _index_searcher = StringField("index_searcher")
    _index_searcher_params = DictField("index_searcher_params")
    _index_reformer = StringField("index_reformer")
    _index_reformer_params = DictField("index_reformer_params")
    _download_index = BoolField("download_index")
    _storage_options = BytesField(
        "storage_options", on_serialize=pickle.dumps, on_deserialize=pickle.loads
    )

    def __init__(
        self,
        tensor=None,
        distance_metric=None,
        dimension=None,
        row_number=None,
        topk=None,
        index=None,
        threads=None,
        index_searcher=None,
        index_searcher_params=None,
        index_reformer=None,
        index_reformer_params=None,
        download_index=None,
        storage_options=None,
        output_types=None,
        stage=None,
        **kw,
    ):
        super().__init__(
            _tensor=tensor,
            _distance_metric=distance_metric,
            _row_number=row_number,
            _dimension=dimension,
            _index=index,
            _threads=threads,
            _index_searcher=index_searcher,
            _index_searcher_params=index_searcher_params,
            _index_reformer=index_reformer,
            _index_reformer_params=index_reformer_params,
            _download_index=download_index,
            _output_types=output_types,
            _topk=topk,
            _storage_options=storage_options,
            **kw,
        )
        if self._output_types is None:
            self._output_types = [OutputType.tensor, OutputType.tensor]

    @property
    def tensor(self):
        return self._tensor

    @property
    def distance_metric(self):
        return self._distance_metric

    @property
    def dimension(self):
        return self._dimension

    @property
    def row_number(self):
        return self._row_number

    @property
    def topk(self):
        return self._topk

    @property
    def threads(self):
        return self._threads

    @property
    def index(self):
        return self._index

    @property
    def index_searcher(self):
        return self._index_searcher

    @property
    def index_searcher_params(self):
        return self._index_searcher_params

    @property
    def index_reformer(self):
        return self._index_reformer

    @property
    def index_reformer_params(self):
        return self._index_reformer_params

    @property
    def download_index(self):
        return self._download_index

    @property
    def storage_options(self):
        return self._storage_options

    @property
    def output_limit(self):
        return 1 if self._download_index else 2

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self.stage != OperandStage.agg and not self._download_index:
            self._tensor = self._inputs[0]
            if isinstance(self._index, ENTITY_TYPE):
                self._index = self._inputs[-1]

    def __call__(self, tensor, index):
        kws = [
            {
                "dtype": np.dtype(np.uint64),
                "shape": (tensor.shape[0], self._topk),
                "order": TensorOrder.C_ORDER,
            },
            {
                "dtype": np.dtype(np.float32),
                "shape": (tensor.shape[0], self._topk),
                "order": TensorOrder.C_ORDER,
            },
        ]
        inputs = [tensor]
        if hasattr(index, "op"):
            inputs.append(index)
        return mt.ExecutableTuple(self.new_tileables(inputs, kws=kws))

    @classmethod
    def _build_download_chunks(cls, op, indexes):
        ctx = get_context()
        workers = ctx.get_worker_addresses() or [None]
        if len(workers) < len(indexes):
            workers = [workers[i % len(workers)] for i in range(len(indexes))]
        indexes_iter = iter(itertools.cycle(indexes))

        download_chunks = defaultdict(list)
        for i, worker in enumerate(workers):
            download_op = op.copy().reset_key()
            download_op.stage = OperandStage.map
            download_op.expect_worker = worker
            download_op._download_index = True
            download_op._tensor = None
            download_op._index = next(indexes_iter)
            download_chunks[i % len(indexes)].append(
                download_op.new_chunk(
                    None, index=(i,), shape=(), dtype=op.inputs[0].dtype
                )
            )
        return download_chunks

    @classmethod
    def tile(cls, op: "ProximaSearcher"):
        tensor = op.tensor
        index = op.index
        topk = op.topk
        outs = op.outputs
        row_number = op.row_number

        ctx = get_context()

        # make sure all inputs have known chunk sizes
        if has_unknown_shape(*op.inputs):
            yield

        rechunk_size = dict()
        if tensor.chunk_shape[1] > 1:
            rechunk_size[1] = tensor.shape[1]
        if row_number is not None:
            rechunk_size[0] = tensor.shape[0] // row_number
        if len(rechunk_size) > 0:
            tensor = yield from recursive_tile(tensor.rechunk(rechunk_size))

        logger.warning(f"query chunks count: {len(tensor.chunks)} ")

        if hasattr(index, "op"):
            built_indexes = [index.chunks] * len(tensor.chunks)
        else:
            # index path
            fs: FileSystem = get_fs(index, op.storage_options)
            index_paths = [
                f for f in fs.ls(index) if f.rsplit("/", 1)[-1].startswith("proxima_")
            ]
            download_chunks = cls._build_download_chunks(op, index_paths)
            iters = [iter(itertools.cycle(i)) for i in download_chunks.values()]
            built_indexes = []
            for _ in range(len(tensor.chunks)):
                built_indexes.append([next(it) for it in iters])

        if hasattr(index, "op"):
            index_chunks_workers = [
                m["bands"][0][0]
                for m in ctx.get_chunks_meta(
                    [c.key for c in index.chunks], fields=["bands"]
                )
            ]
        else:
            index_chunks_workers = [None] * len(built_indexes[0])

        out_chunks = [], []
        for i, tensor_chunk in enumerate(tensor.chunks):
            pk_chunks, distance_chunks = [], []
            for j, chunk_index, worker in zip(
                itertools.count(), built_indexes[i], index_chunks_workers
            ):
                chunk_op = op.copy().reset_key()
                chunk_op.stage = OperandStage.map
                if hasattr(index, "op"):
                    chunk_op.expect_worker = worker
                else:
                    chunk_op.expect_worker = chunk_index.op.expect_worker
                chunk_op._index = chunk_index
                chunk_op._tensor = None
                chunk_kws = [
                    {
                        "index": (tensor_chunk.index[0], j),
                        "dtype": outs[0].dtype,
                        "shape": (tensor_chunk.shape[0], topk),
                        "order": TensorOrder.C_ORDER,
                    },
                    {
                        "index": (tensor_chunk.index[0], j),
                        "dtype": outs[1].dtype,
                        "shape": (tensor_chunk.shape[0], topk),
                        "order": TensorOrder.C_ORDER,
                    },
                ]
                chunk_inputs = [tensor_chunk, chunk_index]
                pk_chunk, distance_chunk = chunk_op.new_chunks(
                    chunk_inputs, kws=chunk_kws
                )
                pk_chunks.append(pk_chunk)
                distance_chunks.append(distance_chunk)

            if len(pk_chunks) == 1:
                out_chunks[0].append(pk_chunks[0])
                out_chunks[1].append(distance_chunks[0])
                continue

            # combine topk results
            combine_size = options.combine_size

            tensor_out_chunks = [pk_chunks, distance_chunks]
            while True:
                chunk_size = ceildiv(len(tensor_out_chunks[0]), combine_size)
                cur_out_chunks = [[], []]
                for k in range(chunk_size):
                    to_combine_pks = tensor_out_chunks[0][
                        k * combine_size : (k + 1) * combine_size
                    ]
                    to_combine_distances = tensor_out_chunks[1][
                        k * combine_size : (k + 1) * combine_size
                    ]

                    chunk_op = op.copy().reset_key()
                    chunk_op.stage = OperandStage.agg
                    chunk_op._tensor = None
                    chunk_op._index = None
                    agg_chunk_kws = [
                        {
                            "index": (i, 0),
                            "dtype": outs[0].dtype,
                            "shape": (tensor_chunk.shape[0], topk),
                            "order": outs[0].order,
                        },
                        {
                            "index": (i, 0),
                            "dtype": outs[1].dtype,
                            "shape": (tensor_chunk.shape[0], topk),
                            "order": outs[1].order,
                        },
                    ]
                    pk_result_chunk, distance_result_chunk = chunk_op.new_chunks(
                        to_combine_pks + to_combine_distances, kws=agg_chunk_kws
                    )
                    cur_out_chunks[0].append(pk_result_chunk)
                    cur_out_chunks[1].append(distance_result_chunk)
                tensor_out_chunks = cur_out_chunks
                if len(tensor_out_chunks[0]) == 1:
                    break
            out_chunks[0].append(tensor_out_chunks[0][0])
            out_chunks[1].append(tensor_out_chunks[1][0])

        kws = []
        pk_params = outs[0].params
        pk_params["chunks"] = out_chunks[0]
        pk_params["nsplits"] = (tensor.nsplits[0], (topk,))
        kws.append(pk_params)
        distance_params = outs[1].params
        distance_params["chunks"] = out_chunks[1]
        distance_params["nsplits"] = (tensor.nsplits[0], (topk,))
        kws.append(distance_params)
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=kws)

    @classmethod
    def _execute_download(cls, ctx, op: "ProximaSearcher"):
        index_path = op.index
        with Timer() as timer:
            fs = get_fs(index_path, op.storage_options)

            # TODO
            dirs = os.environ.get("MARS_SPILL_DIRS")
            if dirs:
                temp_dir = random.choice(dirs.split(":"))
            else:
                temp_dir = "/tmp/proxima-index/"

            local_path = os.path.join(
                temp_dir, md5(str(index_path).encode("utf-8")).hexdigest()
            )  # noqa: B303  # nosec
            exist_state = True
            if not os.path.exists(local_path):
                exist_state = False
                if not os.path.exists(local_path.rsplit("/", 1)[0]):
                    os.mkdir(local_path.rsplit("/", 1)[0])
                with open(local_path, "wb") as out_f:
                    with fs.open(index_path, "rb") as in_f:
                        # 32M
                        chunk_bytes = 32 * 1024**2
                        while True:
                            data = in_f.read(chunk_bytes)
                            if data:
                                out_f.write(data)
                            else:
                                break

        logger.warning(
            f"ReadingFromVolume({op.key}), index path: {index_path}, "
            f"local_path {local_path}"
            f"size {os.path.getsize(local_path)}, "
            f"already exist {exist_state}, "
            f"costs {timer.duration} seconds "
            f"speed {round(os.path.getsize(local_path) / (1024 ** 2) / timer.duration, 2)} MB/s"
        )
        ctx[op.outputs[0].key] = local_path

    @classmethod
    def _execute_map(cls, ctx, op: "ProximaSearcher"):
        if op.download_index:
            cls._execute_download(ctx, op)
            return

        inp = ctx[op.tensor.key]
        index_path = ctx[op.inputs[-1].key]

        with Timer() as timer:
            flow = proxima.IndexFlow(
                container_name="MMapFileContainer",
                container_params={},
                searcher_name=op.index_searcher,
                searcher_params=op.index_searcher_params,
                measure_name="",
                measure_params={},
                reformer_name=op.index_reformer,
                reformer_params=op.index_reformer_params,
            )

            flow.load(index_path)
            vecs = np.ascontiguousarray(inp)

        logger.warning(
            f"LoadIndex({op.key})  index path: {index_path}  costs {timer.duration} seconds"
        )
        logger.warning(f"threads count:{op.threads}  vecs count:{len(vecs)}")

        with Timer() as timer:
            batch = 10000
            s_idx = 0
            e_idx = min(s_idx + batch, len(vecs))
            result_pks, result_distances = None, None
            while s_idx < len(vecs):
                with Timer() as timer_s:
                    tp = get_proxima_type(vecs.dtype)
                    result_pks_b, result_distances_b = proxima.IndexUtility.ann_search(
                        searcher=flow,
                        type=tp,
                        query=vecs[s_idx:e_idx],
                        topk=op.topk,
                        threads=op.threads,
                    )
                    if result_pks is None:
                        result_pks = np.asarray(result_pks_b)
                        result_distances = np.asarray(result_distances_b)
                    else:
                        result_pks = np.concatenate(
                            (result_pks, np.asarray(result_pks_b))
                        )
                        result_distances = np.concatenate(
                            (result_distances, np.asarray(result_distances_b))
                        )

                    s_idx = e_idx
                    e_idx = min(s_idx + batch, len(vecs))
                logger.warning(
                    f"Search({op.key}) count {s_idx}/{len(vecs)}:{round(s_idx * 100 / len(vecs), 2)}%"
                    f" costs {round(timer_s.duration, 2)} seconds"
                )
        logger.warning(f"Search({op.key}) costs {timer.duration} seconds")

        ctx[op.outputs[0].key] = np.asarray(result_pks)
        ctx[op.outputs[1].key] = np.asarray(result_distances)

    @classmethod
    def _execute_agg(cls, ctx, op: "ProximaSearcher"):
        inputs_data = [ctx[inp.key] for inp in op.inputs]

        chunk_num = len(inputs_data) // 2
        pks = np.concatenate(inputs_data[:chunk_num], axis=1)
        distances = np.concatenate(inputs_data[chunk_num:], axis=1)

        n_doc = len(pks)
        topk = op.topk

        # calculate topk on rows
        if op.distance_metric == "InnerProduct":
            inds = np.argsort(distances, axis=1)[:, -1 : -topk - 1 : -1]
        else:
            inds = np.argsort(distances, axis=1)[:, :topk]

        result_pks = np.empty((n_doc, topk), dtype=pks.dtype)
        result_distances = np.empty((n_doc, topk), dtype=distances.dtype)
        rng = np.arange(n_doc)
        for i in range(topk):
            ind = inds[:, i]
            result_pks[:, i] = pks[rng, ind]
            result_distances[:, i] = distances[rng, ind]
        del rng

        ctx[op.outputs[0].key] = result_pks
        ctx[op.outputs[1].key] = result_distances

    @classmethod
    def execute(cls, ctx, op: "ProximaSearcher"):
        if op.stage != OperandStage.agg:
            return cls._execute_map(ctx, op)
        else:
            return cls._execute_agg(ctx, op)


def search_index(
    tensor,
    topk,
    index,
    threads=4,
    row_number=None,
    dimension=None,
    distance_metric=None,
    index_searcher=None,
    index_searcher_params=None,
    index_reformer=None,
    index_reformer_params=None,
    storage_options=None,
    run=True,
    session=None,
    run_kwargs=None,
):
    tensor = validate_tensor(tensor)

    if dimension is None:
        dimension = tensor.shape[1]
    if index_searcher is None:
        index_searcher = ""
    if index_searcher_params is None:
        index_searcher_params = {}
    if index_reformer is None:
        index_reformer = ""
    if index_reformer_params is None:
        index_reformer_params = {}
    if distance_metric is None:
        distance_metric = ""
    if hasattr(index, "op") and index.op.index_path is not None:
        storage_options = storage_options or index.op.storage_options
        index = index.op.index_path

    op = ProximaSearcher(
        tensor=tensor,
        distance_metric=distance_metric,
        dimension=dimension,
        row_number=row_number,
        topk=topk,
        index=index,
        threads=threads,
        index_searcher=index_searcher,
        index_searcher_params=index_searcher_params,
        index_reformer=index_reformer,
        index_reformer_params=index_reformer_params,
        storage_options=storage_options,
    )
    result = op(tensor, index)
    if run:
        return result.execute(session=session, **(run_kwargs or dict()))
    else:
        return result



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import threading
from typing import Dict, Optional


class Isolation:
    loop: asyncio.AbstractEventLoop
    _stopped: Optional[asyncio.Event]
    _thread: Optional[threading.Thread]

    def __init__(self, loop: asyncio.AbstractEventLoop, threaded: bool = True):
        self.loop = loop
        self._threaded = threaded

        self._stopped = None
        self._thread = None
        self._thread_ident = None

    def _run(self):
        asyncio.set_event_loop(self.loop)
        self._stopped = asyncio.Event()
        self.loop.run_until_complete(self._stopped.wait())

    def start(self):
        if self._threaded:
            self._thread = thread = threading.Thread(target=self._run)
            thread.daemon = True
            thread.start()
            self._thread_ident = thread.ident

    @property
    def thread_ident(self):
        return self._thread_ident

    async def _stop(self):
        self._stopped.set()

    def stop(self):
        if self._threaded:
            asyncio.run_coroutine_threadsafe(self._stop(), self.loop).result()
            self._thread.join()


_name_to_isolation: Dict[str, Isolation] = dict()


DEFAULT_ISOLATION = "oscar"


def new_isolation(
    name: str = DEFAULT_ISOLATION,
    loop: asyncio.AbstractEventLoop = None,
    threaded: bool = True,
) -> Isolation:
    if name in _name_to_isolation:
        return _name_to_isolation[name]

    if loop is None:
        loop = asyncio.new_event_loop()

    isolation = Isolation(loop, threaded=threaded)
    isolation.start()
    _name_to_isolation[name] = isolation
    return isolation


def get_isolation(name: str = DEFAULT_ISOLATION):
    isolation = _name_to_isolation[name]
    if isolation.loop.is_closed():  # pragma: no cover
        _name_to_isolation.pop(name)
        raise KeyError(name)
    return isolation


def stop_isolation(name: str = DEFAULT_ISOLATION):
    if name in _name_to_isolation:
        return _name_to_isolation.pop(name).stop()



# The MIT License
#
# Copyright (c) 2018 aio-libs team https://github.com/aio-libs/
# Copyright (c) 2017 Ocean S. A. https://ocean.io/
# Copyright (c) 2016-2017 WikiBusiness Corporation http://wikibusiness.org/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import asyncio
import os
import weakref
from collections import OrderedDict
from functools import _CacheInfo, _make_key, partial, wraps


__version__ = "1.0.2"

__all__ = ("alru_cache", "clear_all_alru_caches")

_is_ci = (os.environ.get("CI") or "0").lower() in ("1", "true")
_all_wrapped = weakref.WeakSet()


def clear_all_alru_caches():
    for wrapped in _all_wrapped:
        wrapped.cache_clear()


def unpartial(fn):
    while hasattr(fn, "func"):
        fn = fn.func

    return fn


def _done_callback(fut, task):
    if task.cancelled():
        fut.cancel()
        return

    exc = task.exception()
    if exc is not None:
        fut.set_exception(exc)
        return

    fut.set_result(task.result())


def _cache_invalidate(wrapped, typed, *args, **kwargs):
    key = _make_key(args, kwargs, typed)

    exists = key in wrapped._cache

    if exists:
        wrapped._cache.pop(key)

    return exists


def _cache_clear(wrapped):
    wrapped.hits = wrapped.misses = 0
    wrapped._cache = OrderedDict()
    wrapped.tasks = set()


def _open(wrapped):
    if not wrapped.closed:
        raise RuntimeError("alru_cache is not closed")

    was_closed = (
        wrapped.hits == wrapped.misses == len(wrapped.tasks) == len(wrapped._cache) == 0
    )

    if not was_closed:
        raise RuntimeError("alru_cache was not closed correctly")

    wrapped.closed = False


def _close(wrapped, *, cancel=False, return_exceptions=True):
    if wrapped.closed:
        raise RuntimeError("alru_cache is closed")

    wrapped.closed = True

    if cancel:
        for task in wrapped.tasks:
            if not task.done():  # not sure is it possible
                task.cancel()

    return _wait_closed(wrapped, return_exceptions=return_exceptions)


async def _wait_closed(wrapped, *, return_exceptions):
    wait_closed = asyncio.gather(*wrapped.tasks, return_exceptions=return_exceptions)

    wait_closed.add_done_callback(partial(_close_waited, wrapped))

    ret = await wait_closed

    # hack to get _close_waited callback to be executed
    await asyncio.sleep(0)

    return ret


def _close_waited(wrapped, _):
    wrapped.cache_clear()


def _cache_info(wrapped, maxsize):
    return _CacheInfo(
        wrapped.hits,
        wrapped.misses,
        maxsize,
        len(wrapped._cache),
    )


def __cache_touch(wrapped, key):
    try:
        wrapped._cache.move_to_end(key)
    except KeyError:  # not sure is it possible
        pass


def _cache_hit(wrapped, key):
    wrapped.hits += 1
    __cache_touch(wrapped, key)


def _cache_miss(wrapped, key):
    wrapped.misses += 1
    __cache_touch(wrapped, key)


def alru_cache(
    fn=None,
    maxsize=128,
    typed=False,
    *,
    cache_exceptions=True,
):
    def wrapper(fn):
        _origin = unpartial(fn)

        if not asyncio.iscoroutinefunction(_origin):
            raise RuntimeError("Coroutine function is required, got {}".format(fn))

        # functools.partialmethod support
        if hasattr(fn, "_make_unbound_method"):
            fn = fn._make_unbound_method()

        @wraps(fn)
        async def wrapped(*fn_args, **fn_kwargs):
            if wrapped.closed:
                raise RuntimeError("alru_cache is closed for {}".format(wrapped))

            loop = asyncio.get_event_loop()

            key = _make_key(fn_args, fn_kwargs, typed)

            fut = wrapped._cache.get(key)

            if fut is not None:
                if not fut.done():
                    _cache_hit(wrapped, key)
                    return await asyncio.shield(fut)

                exc = fut._exception

                if exc is None or cache_exceptions:
                    _cache_hit(wrapped, key)
                    return fut.result()

                # exception here and cache_exceptions == False
                wrapped._cache.pop(key)

            fut = loop.create_future()
            task = loop.create_task(fn(*fn_args, **fn_kwargs))
            task.add_done_callback(partial(_done_callback, fut))

            wrapped.tasks.add(task)
            task.add_done_callback(wrapped.tasks.remove)

            wrapped._cache[key] = fut

            if maxsize is not None and len(wrapped._cache) > maxsize:
                wrapped._cache.popitem(last=False)

            _cache_miss(wrapped, key)
            return await asyncio.shield(fut)

        _cache_clear(wrapped)
        wrapped._origin = _origin
        wrapped.closed = False
        wrapped.cache_info = partial(_cache_info, wrapped, maxsize)
        wrapped.cache_clear = partial(_cache_clear, wrapped)
        wrapped.invalidate = partial(_cache_invalidate, wrapped, typed)
        wrapped.close = partial(_close, wrapped)
        wrapped.open = partial(_open, wrapped)

        if _is_ci:
            _all_wrapped.add(wrapped)
        return wrapped

    if fn is None:
        return wrapper

    if callable(fn) or hasattr(fn, "_make_unbound_method"):
        return wrapper(fn)

    raise NotImplementedError("{} decorating is not supported".format(fn))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import time
from typing import Callable, Dict, List, Optional, Set

from ....lib.aio import alru_cache
from ....resource import Resource
from ....typing import BandType
from ....utils import serialize_serializable, deserialize_serializable
from ...core import NodeRole
from ...web import web_api, MarsServiceWebAPIHandler, MarsWebAPIClientMixin
from ..core import watch_method, NodeStatus
from .core import AbstractClusterAPI


class ClusterWebAPIHandler(MarsServiceWebAPIHandler):
    _root_pattern = "/api/cluster"

    @alru_cache(cache_exceptions=False)
    async def _get_cluster_api(self):
        from ...cluster import ClusterAPI

        return await ClusterAPI.create(self._supervisor_addr)

    @staticmethod
    def _convert_node_dict(node_info_list: Dict[str, Dict]):
        res = {}
        for node_addr, node in node_info_list.items():
            res_dict = node.copy()
            res_dict["status"] = res_dict["status"].value
            res[node_addr] = res_dict
        return res

    @web_api("nodes", method=["get", "post"], cache_blocking=True)
    async def get_nodes_info(self):
        watch = bool(int(self.get_argument("watch", "0")))
        env = bool(int(self.get_argument("env", "0")))
        resource = bool(int(self.get_argument("resource", "0")))
        detail = bool(int(self.get_argument("detail", "0")))

        nodes_arg = self.get_argument("nodes", None)
        nodes = nodes_arg.split(",") if nodes_arg is not None else None

        role_arg = self.get_argument("role", None)
        role = NodeRole(int(role_arg)) if role_arg is not None else None

        statuses_arg = self.get_argument("statuses", None)
        statuses = (
            set(NodeStatus(int(v)) for v in statuses_arg.split(","))
            if statuses_arg
            else None
        )

        exclude_statuses_arg = self.get_argument("exclude_statuses", None)
        exclude_statuses = (
            set(NodeStatus(int(v)) for v in exclude_statuses_arg.split(","))
            if exclude_statuses_arg
            else None
        )

        statuses = WebClusterAPI._calc_statuses(statuses, exclude_statuses)

        cluster_api = await self._get_cluster_api()
        result = {}
        if watch:
            assert nodes is None
            version = self.get_argument("version", "") or None
            if version:
                version = int(version)

            async for version, node_infos in cluster_api.watch_nodes(
                role,
                env=env,
                resource=resource,
                detail=detail,
                statuses=statuses,
                version=version,
            ):
                result["version"] = version
                result["nodes"] = self._convert_node_dict(node_infos)
                break
        else:
            nodes = await cluster_api.get_nodes_info(
                nodes=nodes,
                role=role,
                env=env,
                resource=resource,
                statuses=statuses,
                detail=detail,
            )
            result["nodes"] = self._convert_node_dict(nodes)
        self.write(json.dumps(result))

    @web_api("bands", method="get", cache_blocking=True)
    async def get_all_bands(self):
        role_arg = self.get_argument("role", None)
        role = NodeRole(int(role_arg)) if role_arg is not None else None
        watch = bool(int(self.get_argument("watch", "0")))

        statuses_arg = self.get_argument("statuses", None)
        statuses = (
            set(NodeStatus(int(v)) for v in statuses_arg.split(","))
            if statuses_arg
            else None
        )

        cluster_api = await self._get_cluster_api()
        if watch:
            version = self.get_argument("version", "") or None
            if version:
                version = int(version)

            async for version, bands in cluster_api.watch_all_bands(
                role, statuses=statuses, version=version
            ):
                self.write(serialize_serializable((version, bands)))
                break
        else:
            self.write(
                serialize_serializable(
                    await cluster_api.get_all_bands(role, statuses=statuses)
                )
            )

    @web_api("versions", method="get", cache_blocking=True)
    async def get_mars_versions(self):
        cluster_api = await self._get_cluster_api()
        self.write(json.dumps(list(await cluster_api.get_mars_versions())))

    @web_api("pools", method="get", cache_blocking=True)
    async def get_node_pool_configs(self):
        cluster_api = await self._get_cluster_api()
        address = self.get_argument("address", "") or None
        pools = list(await cluster_api.get_node_pool_configs(address))
        self.write(json.dumps({"pools": pools}))

    @web_api("stacks", method="get", cache_blocking=True)
    async def get_node_thread_stacks(self):
        cluster_api = await self._get_cluster_api()
        address = self.get_argument("address", "") or None
        stacks = list(await cluster_api.get_node_thread_stacks(address))
        self.write(
            json.dumps(
                {
                    "generate_time": time.time(),
                    "stacks": stacks,
                }
            )
        )


web_handlers = {ClusterWebAPIHandler.get_root_pattern(): ClusterWebAPIHandler}


class WebClusterAPI(AbstractClusterAPI, MarsWebAPIClientMixin):
    def __init__(self, address: str, request_rewriter: Callable = None):
        self._address = address.rstrip("/")
        self.request_rewriter = request_rewriter

    @staticmethod
    def _convert_node_dict(node_info_list: Dict[str, Dict]):
        res = {}
        for node_addr, node in node_info_list.items():
            res_dict = node.copy()
            res_dict["status"] = NodeStatus(res_dict["status"])
            res[node_addr] = res_dict
        return res

    async def _get_nodes_info(
        self,
        nodes: List[str] = None,
        role: NodeRole = None,
        env: bool = False,
        resource: bool = False,
        detail: bool = False,
        watch: bool = False,
        statuses: Set[NodeStatus] = None,
        version: Optional[int] = None,
    ):
        statuses_str = (
            ",".join(str(status.value) for status in statuses) if statuses else ""
        )
        args = [
            ("nodes", ",".join(nodes) if nodes else None),
            ("role", role.value if role is not None else None),
            ("env", 1 if env else 0),
            ("resource", 1 if resource else 0),
            ("detail", 1 if detail else 0),
            ("watch", 1 if watch else 0),
            ("statuses", statuses_str),
            ("version", str(version or "")),
        ]
        args_str = "&".join(f"{key}={val}" for key, val in args if val is not None)

        path = f"{self._address}/api/cluster/nodes"
        res = await self._request_url(
            path=path,
            method="POST",
            data=args_str,
            headers={"Content-Type": "application/x-www-form-urlencoded"},
        )
        result = json.loads(res.body)
        if watch:
            return result["version"], self._convert_node_dict(result["nodes"])
        else:
            return self._convert_node_dict(result["nodes"])

    async def get_supervisors(self, filter_ready: bool = True) -> List[str]:
        statuses = (
            {NodeStatus.READY}
            if filter_ready
            else {NodeStatus.STARTING, NodeStatus.READY}
        )
        res = await self._get_nodes_info(role=NodeRole.SUPERVISOR, statuses=statuses)
        return list(res.keys())

    @watch_method
    async def watch_supervisors(self, version: Optional[int] = None):
        version, res = await self._get_nodes_info(
            role=NodeRole.SUPERVISOR, watch=True, version=version
        )
        return version, list(res.keys())

    async def get_nodes_info(
        self,
        nodes: List[str] = None,
        role: NodeRole = None,
        env: bool = False,
        resource: bool = False,
        detail: bool = False,
        statuses: Set[NodeStatus] = None,
        exclude_statuses: Set[NodeStatus] = None,
    ):
        statuses = self._calc_statuses(statuses, exclude_statuses)
        return await self._get_nodes_info(
            nodes,
            role=role,
            env=env,
            resource=resource,
            detail=detail,
            watch=False,
            statuses=statuses,
        )

    @watch_method
    async def watch_nodes(
        self,
        role: NodeRole,
        env: bool = False,
        resource: bool = False,
        detail: bool = False,
        statuses: Set[NodeStatus] = None,
        exclude_statuses: Set[NodeStatus] = None,
        version: Optional[int] = None,
    ) -> List[Dict[str, Dict]]:
        statuses = self._calc_statuses(statuses, exclude_statuses)
        return await self._get_nodes_info(
            role=role,
            env=env,
            resource=resource,
            detail=detail,
            watch=True,
            statuses=statuses,
            version=version,
        )

    async def get_all_bands(
        self,
        role: NodeRole = None,
        statuses: Set[NodeStatus] = None,
        exclude_statuses: Set[NodeStatus] = None,
    ) -> Dict[BandType, Resource]:
        statuses = self._calc_statuses(statuses, exclude_statuses)
        statuses_str = (
            ",".join(str(status.value) for status in statuses) if statuses else ""
        )
        params = {}
        if role is not None:  # pragma: no cover
            params["role"] = role.value
        if statuses_str:
            params["statuses"] = statuses_str

        path = f"{self._address}/api/cluster/bands"
        res = await self._request_url("GET", path, params=params)
        return deserialize_serializable(res.body)

    @watch_method
    async def watch_all_bands(
        self,
        role: NodeRole = None,
        statuses: List[NodeStatus] = None,
        exclude_statuses: Set[NodeStatus] = None,
        version: Optional[int] = None,
    ):
        statuses = self._calc_statuses(statuses, exclude_statuses)
        statuses_str = (
            ",".join(str(status.value) for status in statuses) if statuses else ""
        )
        params = dict(watch=1, version=str(version or ""))
        if role is not None:  # pragma: no cover
            params["role"] = role.value
        if statuses_str:
            params["statuses"] = statuses_str

        path = f"{self._address}/api/cluster/bands"
        res = await self._request_url("GET", path, params=params)
        return deserialize_serializable(res.body)

    async def get_mars_versions(self) -> List[str]:
        path = f"{self._address}/api/cluster/versions"
        res = await self._request_url("GET", path)
        return list(json.loads(res.body))

    async def get_node_pool_configs(self, address: str) -> List[Dict]:
        path = f"{self._address}/api/cluster/pools?address={address}"
        res = await self._request_url("GET", path)
        return list(json.loads(res.body)["pools"])

    async def get_node_thread_stacks(self, address: str) -> List[Dict]:
        path = f"{self._address}/api/cluster/stacks?address={address}"
        res = await self._request_url("GET", path)
        return list(json.loads(res.body)["stacks"])



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .pairwise import euclidean_distances, pairwise_distances, pairwise_distances_topk
from ._classification import (
    accuracy_score,
    log_loss,
    multilabel_confusion_matrix,
    precision_recall_fscore_support,
    precision_score,
    recall_score,
    f1_score,
    fbeta_score,
)
from ._ranking import roc_curve, auc, roc_auc_score
from ._regresssion import r2_score
from ._scorer import get_scorer



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .loc import DataFrameLoc


class DataFrameAt:
    def __init__(self, obj):
        self._obj = obj
        self._loc = DataFrameLoc(self._obj)

    def __getitem__(self, indexes):
        if not isinstance(indexes, tuple):
            indexes = (indexes,)

        for index in indexes:
            if not np.isscalar(index):
                raise ValueError("Invalid call for scalar access (getting)!")

        return self._loc[indexes]


def at(a):
    """
    Access a single value for a row/column label pair.

    Similar to ``loc``, in that both provide label-based lookups. Use
    ``at`` if you only need to get or set a single value in a DataFrame
    or Series.

    Raises
    ------
    KeyError
        If 'label' does not exist in DataFrame.

    See Also
    --------
    DataFrame.iat : Access a single value for a row/column pair by integer
        position.
    DataFrame.loc : Access a group of rows and columns by label(s).
    Series.at : Access a single value using a label.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame([[0, 2, 3], [0, 4, 1], [10, 20, 30]],
    ...                   index=[4, 5, 6], columns=['A', 'B', 'C'])
    >>> df.execute()
        A   B   C
    4   0   2   3
    5   0   4   1
    6  10  20  30

    Get value at specified row/column pair

    >>> df.at[4, 'B'].execute()
    2

    # Set value at specified row/column pair
    #
    # >>> df.at[4, 'B'] = 10
    # >>> df.at[4, 'B']
    # 10

    Get value within a Series

    >>> df.loc[5].at['B'].execute()
    4
    """
    return DataFrameAt(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy.linalg import LinAlgError

from ... import opcodes as OperandDef
from ...serialization.serializables import KeyField, StringField
from ...core import ExecutableTuple
from ..array_utils import as_same_device, device
from ..datasource import tensor as astensor
from ..operands import TensorHasInput, TensorOperandMixin
from ..core import TensorOrder
from .core import TSQR
from .utils import calc_svd_shapes


class TensorSVD(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.SVD

    _input = KeyField("input")
    _method = StringField("method")

    def __init__(self, method=None, **kw):
        super().__init__(_method=method, **kw)

    @property
    def method(self):
        return self._method

    @property
    def output_limit(self):
        return 3

    @classmethod
    def _is_svd(cls):
        return True

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, a):
        a = astensor(a)

        if a.ndim != 2:
            raise LinAlgError(
                f"{a.ndim}-dimensional tensor given. Tensor must be two-dimensional"
            )

        tiny_U, tiny_s, tiny_V = np.linalg.svd(np.ones((1, 1), dtype=a.dtype))

        # if a's shape is (6, 18), U's shape is (6, 6), s's shape is (6,), V's shape is (6, 18)
        # if a's shape is (18, 6), U's shape is (18, 6), s's shape is (6,), V's shape is (6, 6)
        U_shape, s_shape, V_shape = calc_svd_shapes(a)
        U, s, V = self.new_tensors(
            [a],
            order=TensorOrder.C_ORDER,
            kws=[
                {"side": "U", "dtype": tiny_U.dtype, "shape": U_shape},
                {"side": "s", "dtype": tiny_s.dtype, "shape": s_shape},
                {"side": "V", "dtype": tiny_V.dtype, "shape": V_shape},
            ],
        )
        return ExecutableTuple([U, s, V])

    @classmethod
    def tile(cls, op):
        U, s, V = op.outputs
        U_dtype, s_dtype, V_dtype = U.dtype, s.dtype, V.dtype
        U_shape, s_shape, V_shape = U.shape, s.shape, V.shape
        in_tensor = op.input
        if in_tensor.chunk_shape == (1, 1):
            in_chunk = in_tensor.chunks[0]
            chunk_op = op.copy().reset_key()
            svd_chunks = chunk_op.new_chunks(
                [in_chunk],
                kws=[
                    {
                        "side": "U",
                        "dtype": U_dtype,
                        "index": in_chunk.index,
                        "shape": U_shape,
                        "order": U.order,
                    },
                    {
                        "side": "s",
                        "dtype": s_dtype,
                        "index": in_chunk.index[1:],
                        "shape": s_shape,
                        "order": s.order,
                    },
                    {
                        "side": "V",
                        "dtype": V_dtype,
                        "index": in_chunk.index,
                        "shape": V_shape,
                        "order": V.order,
                    },
                ],
            )
            U_chunk, s_chunk, V_chunk = svd_chunks

            new_op = op.copy()
            kws = [
                {
                    "chunks": [U_chunk],
                    "nsplits": tuple((s,) for s in U_shape),
                    "dtype": U_dtype,
                    "shape": U_shape,
                },
                {
                    "chunks": [s_chunk],
                    "nsplits": tuple((s,) for s in s_shape),
                    "dtype": s_dtype,
                    "shape": s_shape,
                },
                {
                    "chunks": [V_chunk],
                    "nsplits": tuple((s,) for s in V_shape),
                    "dtype": V_dtype,
                    "shape": V_shape,
                },
            ]
            return new_op.new_tensors(op.inputs, kws=kws)
        elif op.method == "tsqr":
            return (yield from TSQR.tile(op))
        else:
            raise NotImplementedError("Only tsqr method supported for now")

    @classmethod
    def execute(cls, ctx, op):
        (a,), device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )

        with device(device_id):
            u, s, v = xp.linalg.svd(a, full_matrices=False)
            uc, sc, vc = op.outputs
            ctx[uc.key] = u
            ctx[sc.key] = s
            ctx[vc.key] = v


def svd(a, method="tsqr"):
    """
    Singular Value Decomposition.

    When `a` is a 2D tensor, it is factorized as ``u @ np.diag(s) @ vh
    = (u * s) @ vh``, where `u` and `vh` are 2D unitary tensors and `s` is a 1D
    tensor of `a`'s singular values. When `a` is higher-dimensional, SVD is
    applied in stacked mode as explained below.

    Parameters
    ----------
    a : (..., M, N) array_like
        A real or complex tensor with ``a.ndim >= 2``.
    method: {'tsqr'}, optional
        method to calculate qr factorization, tsqr as default

        TSQR is presented in:

            A. Benson, D. Gleich, and J. Demmel.
            Direct QR factorizations for tall-and-skinny matrices in
            MapReduce architectures.
            IEEE International Conference on Big Data, 2013.
            http://arxiv.org/abs/1301.1071


    Returns
    -------
    u : { (..., M, M), (..., M, K) } tensor
        Unitary tensor(s). The first ``a.ndim - 2`` dimensions have the same
        size as those of the input `a`. The size of the last two dimensions
        depends on the value of `full_matrices`. Only returned when
        `compute_uv` is True.
    s : (..., K) tensor
        Vector(s) with the singular values, within each vector sorted in
        descending order. The first ``a.ndim - 2`` dimensions have the same
        size as those of the input `a`.
    vh : { (..., N, N), (..., K, N) } tensor
        Unitary tensor(s). The first ``a.ndim - 2`` dimensions have the same
        size as those of the input `a`. The size of the last two dimensions
        depends on the value of `full_matrices`. Only returned when
        `compute_uv` is True.

    Raises
    ------
    LinAlgError
        If SVD computation does not converge.

    Notes
    -----

    SVD is usually described for the factorization of a 2D matrix :math:`A`.
    The higher-dimensional case will be discussed below. In the 2D case, SVD is
    written as :math:`A = U S V^H`, where :math:`A = a`, :math:`U= u`,
    :math:`S= \\mathtt{np.diag}(s)` and :math:`V^H = vh`. The 1D tensor `s`
    contains the singular values of `a` and `u` and `vh` are unitary. The rows
    of `vh` are the eigenvectors of :math:`A^H A` and the columns of `u` are
    the eigenvectors of :math:`A A^H`. In both cases the corresponding
    (possibly non-zero) eigenvalues are given by ``s**2``.

    If `a` has more than two dimensions, then broadcasting rules apply, as
    explained in :ref:`routines.linalg-broadcasting`. This means that SVD is
    working in "stacked" mode: it iterates over all indices of the first
    ``a.ndim - 2`` dimensions and for each combination SVD is applied to the
    last two indices. The matrix `a` can be reconstructed from the
    decomposition with either ``(u * s[..., None, :]) @ vh`` or
    ``u @ (s[..., None] * vh)``. (The ``@`` operator can be replaced by the
    function ``mt.matmul`` for python versions below 3.5.)

    Examples
    --------
    >>> import mars.tensor as mt
    >>> a = mt.random.randn(9, 6) + 1j*mt.random.randn(9, 6)
    >>> b = mt.random.randn(2, 7, 8, 3) + 1j*mt.random.randn(2, 7, 8, 3)

    Reconstruction based on reduced SVD, 2D case:

    >>> u, s, vh = mt.linalg.svd(a)
    >>> u.shape, s.shape, vh.shape
    ((9, 6), (6,), (6, 6))
    >>> np.allclose(a, np.dot(u * s, vh))
    True
    >>> smat = np.diag(s)
    >>> np.allclose(a, np.dot(u, np.dot(smat, vh)))
    True

    """
    op = TensorSVD(method=method)
    return op(a)



# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import textwrap
from urllib.parse import urlparse

from ...utils import parse_readable_size


def _remove_nones(cfg):
    return dict((k, v) for k, v in cfg.items() if v is not None)


def _get_local_app_module(mod_name):
    return __name__.rsplit(".", 1)[0] + "." + mod_name.rsplit(".", 1)[-1]


class SecurityConfig:
    def __init__(self, cert_file=None, key_file=None):
        self._cert_file = cert_file
        self._key_file = key_file

    def build(self):
        return dict(cert_file=self._cert_file, key_file=self._key_file)


class AppFileConfig:
    def __init__(
        self, source, file_type=None, visibility=None, size=None, timestamp=None
    ):
        self._source = source
        self._file_type = file_type
        self._visibility = visibility
        self._size = size
        self._timestamp = timestamp

    def build(self):
        if all(
            v is None
            for v in (self._file_type, self._visibility, self._size, self._timestamp)
        ):
            return self._source
        else:
            return _remove_nones(
                dict(
                    source=self._source,
                    type=self._file_type,
                    visibility=self._visibility,
                    size=self._size,
                    timestamp=self._timestamp,
                )
            )


class AppContainerConfig:
    def __init__(self, cpu=None, memory=None, env=None, files=None, script=None):
        self._cpu = cpu

        if memory is not None:
            real_mem, is_percent = parse_readable_size(memory)
            assert not is_percent
            self._memory = real_mem
        else:
            self._memory = None

        self._env = env
        self._script = script
        self._files = files

        self.add_default_envs()

    def build_script(self):
        return self._script

    def add_default_envs(self):
        pass

    def add_env(self, k, v):
        if self._env is None:
            self._env = dict()
        self._env[k] = v

    def build(self):
        return _remove_nones(
            dict(
                resources=dict(
                    vcores=self._cpu,
                    memory=f"{self._memory // 1024 ** 2} MiB" if self._memory else None,
                ),
                env=self._env,
                script=self.build_script(),
                files=dict((k, v.build()) for k, v in self._files.items())
                if self._files
                else None,
            )
        )


class AppMasterConfig(AppContainerConfig):
    def __init__(self, security=None, **kwargs):
        super().__init__(**kwargs)
        self._security = security

    def build(self):
        d = super().build()
        if self._security is not None:
            d["security"] = self._security.build()
        return d


class AppServiceConfig(AppContainerConfig):
    def __init__(
        self, instances=1, depends=None, allow_failures=False, max_restarts=0, **kwargs
    ):
        super().__init__(**kwargs)
        if isinstance(depends, str):
            depends = [depends]

        self._allow_failures = allow_failures
        self._depends = depends or []
        self._max_restarts = max_restarts
        self._instances = instances

    def build(self):
        d = super().build()
        d.update(
            dict(
                instances=self._instances,
                depends=self._depends,
                allow_failures=self._allow_failures,
                max_restarts=self._max_restarts,
            )
        )
        return d


class MarsServiceConfig(AppServiceConfig):
    service_name = None

    def __init__(
        self,
        environment,
        modules=None,
        cmd_tmpl=None,
        cpu=None,
        memory=None,
        log_config=None,
        extra_args=None,
        **kwargs,
    ):
        files = kwargs.pop("files", dict())
        kwargs["files"] = files

        parsed = urlparse(environment)
        self._env_scheme = parsed.scheme

        if parsed.scheme:
            import mars

            self._source_path = os.path.dirname(
                os.path.dirname(os.path.abspath(mars.__file__))
            )

            self._env_path = environment[len(parsed.scheme) + 3 :]
            self._path_environ = os.environ["PATH"]
        else:
            self._source_path = None
            self._env_path = environment
            self._path_environ = None

        self._cmd_tmpl = cmd_tmpl or '"{executable}"'
        if not self._env_scheme:
            files["mars_env"] = AppFileConfig(environment)

        self._log_config = log_config
        if log_config:
            files["logging.conf"] = AppFileConfig(log_config)

        self._modules = modules.split(",") if isinstance(modules, str) else modules

        self._extra_args = extra_args or ""

        cpu = cpu or 1
        memory = memory or "1 GiB"
        super().__init__(cpu=cpu, memory=memory, **kwargs)

    def add_default_envs(self):
        if self._cpu:
            self.add_env("MKL_NUM_THREADS", str(self._cpu))
            self.add_env("MARS_CPU_TOTAL", str(self._cpu))
            self.add_env("MARS_USE_PROCESS_STAT", "1")

        if self._memory:
            self.add_env("MARS_MEMORY_TOTAL", str(int(self._memory)))

        if self._modules:
            self.add_env("MARS_LOAD_MODULES", ",".join(self._modules))

        if self._path_environ:
            self.add_env("MARS_YARN_PATH", self._path_environ)

        if self._source_path:
            self.add_env("MARS_SOURCE_PATH", self._source_path)

    def build_script(self):
        bash_lines = [
            textwrap.dedent(
                """
        #!/bin/bash
        if [[ "$YARN_CONTAINER_RUNTIME_TYPE" == "docker" ]]; then
          export MARS_USE_CGROUP_STAT=1
        else
          export MARS_USE_PROCESS_STAT=1
        fi
        if [[ -n $MARS_SOURCE_PATH ]]; then export PYTHONPATH=$PYTHONPATH:$MARS_SOURCE_PATH; fi
        if [[ -n $MARS_YARN_PATH ]]; then export PATH=$MARS_YARN_PATH:$PATH; fi
        """
            ).strip()
        ]

        if not self._env_scheme:
            bash_lines.append("source mars_env/bin/activate")
            python_executable = "mars_env/bin/python"
        elif self._env_scheme == "conda":
            bash_lines.append(f'conda activate "{self._env_path}"')
            python_executable = "python"
        elif self._env_scheme == "venv":
            bash_lines.append(f'source "{self._env_path}/bin/activate"')
            python_executable = self._env_path + "/bin/python"
        else:  # pragma: no cover
            python_executable = self._env_path

        cmd = self._cmd_tmpl.format(executable=python_executable)
        bash_lines.append(
            f"{cmd} -m {_get_local_app_module(self.service_name)} {self._extra_args} > /tmp/{self.service_name}.stdout.log 2> /tmp/{self.service_name}.stderr.log"
        )
        return "\n".join(bash_lines) + "\n"


class MarsSupervisorConfig(MarsServiceConfig):
    service_name = "mars.supervisor"
    web_service_name = "mars.web"


class MarsWorkerConfig(MarsServiceConfig):
    service_name = "mars.worker"

    def __init__(self, environment, worker_cache_mem=None, spill_dirs=None, **kwargs):
        kwargs["depends"] = MarsSupervisorConfig.service_name
        super().__init__(environment, **kwargs)

        if worker_cache_mem:
            self.add_env("MARS_CACHE_MEM_SIZE", worker_cache_mem)

        if spill_dirs:
            self.add_env(
                "MARS_SPILL_DIRS",
                spill_dirs if isinstance(spill_dirs, str) else ":".join(spill_dirs),
            )


class MarsApplicationConfig:
    def __init__(
        self,
        name=None,
        queue=None,
        file_systems=None,
        master=None,
        supervisor_config=None,
        worker_config=None,
    ):
        self._name = name
        self._queue = queue or "default"
        self._file_systems = file_systems or []
        self._master = master or AppMasterConfig(cpu=1, memory="512 MiB")
        self._supervisor_config = supervisor_config
        self._worker_config = worker_config

    def build(self):
        services = _remove_nones(
            {
                MarsSupervisorConfig.service_name: self._supervisor_config.build()
                if self._supervisor_config
                else None,
                MarsWorkerConfig.service_name: self._worker_config.build()
                if self._worker_config
                else None,
            }
        )
        return dict(
            name=self._name,
            queue=self._queue,
            file_systems=self._file_systems,
            master=self._master.build() if self._master else None,
            services=services,
        )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import uuid
from typing import List, Optional, Union

import numpy as np

from ...core import recursive_tile
from ...core.context import get_context, Context
from ...config import options
from ...oscar import ActorNotExist
from ...serialization.serializables import Int64Field, StringField
from ...typing import TileableType, OperandType
from ...utils import parse_readable_size
from ..core import IndexValue, OutputType
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import merge_index_value


class HeadOptimizedDataSource(DataFrameOperand, DataFrameOperandMixin):
    __slots__ = ()
    # Data source op that optimized for head,
    # First, it will try to trigger first_chunk.head() and raise TilesError,
    # When iterative tiling is triggered,
    # check if the first_chunk.head() meets requirements.
    nrows = Int64Field("nrows", default=None)

    @property
    def first_chunk(self):
        return getattr(self, "_first_chunk", None)

    @classmethod
    def _tile(cls, op):  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def _tile_head(cls, op: "HeadOptimizedDataSource"):
        tileds = cls._tile(op)
        chunks = tileds[0].chunks

        # execute first chunk
        yield chunks[:1]

        chunk_shape = chunks[0].shape
        if chunk_shape[0] == op.nrows:
            # the first chunk has enough data
            tileds[0]._nsplits = tuple((s,) for s in chunk_shape)
            chunks[0]._shape = chunk_shape
            tileds[0]._chunks = chunks[:1]
            tileds[0]._shape = chunk_shape
        else:
            for chunk in tileds[0].chunks:
                chunk.op.nrows = None
            # otherwise
            tiled = yield from recursive_tile(tileds[0].iloc[: op.nrows])
            tileds = [tiled]
        return tileds

    @classmethod
    def tile(cls, op: "HeadOptimizedDataSource"):
        if op.nrows is not None:
            return (yield from cls._tile_head(op))
        else:
            return cls._tile(op)


class ColumnPruneSupportedDataSourceMixin(DataFrameOperandMixin):
    __slots__ = ()

    def get_columns(self):  # pragma: no cover
        raise NotImplementedError

    def set_pruned_columns(self, columns, *, keep_order=None):  # pragma: no cover
        raise NotImplementedError


class _IncrementalIndexRecorder:
    _done: List[Optional[asyncio.Event]]
    _chunk_sizes: List[Optional[int]]

    def __init__(self, n_chunk: int):
        self._n_chunk = n_chunk
        self._done = [asyncio.Event() for _ in range(n_chunk)]
        self._chunk_sizes = [None] * n_chunk
        self._waiters = set()

    def _can_destroy(self):
        return all(e.is_set() for e in self._done) and not self._waiters

    def add_waiter(self, i: int):
        self._waiters.add(i)

    async def wait(self, i: int):
        if i == 0:
            return 0, self._can_destroy()
        self._waiters.add(i)
        try:
            await asyncio.gather(*(e.wait() for e in self._done[:i]))
        finally:
            self._waiters.remove(i)
        # all chunk finished and no waiters
        return sum(self._chunk_sizes[:i]), self._can_destroy()

    async def finish(self, i: int, size: int):
        self._chunk_sizes[i] = size
        self._done[i].set()


class IncrementalIndexDatasource(HeadOptimizedDataSource):
    __slots__ = ()

    incremental_index_recorder_name = StringField("incremental_index_recorder_name")


class IncrementalIndexDataSourceMixin(DataFrameOperandMixin):
    __slots__ = ()

    @classmethod
    def post_tile(cls, op: OperandType, results: List[TileableType]):
        if (
            op.incremental_index
            and results is not None
            and isinstance(results[0].index_value.value, IndexValue.RangeIndex)
        ):
            result = results[0]
            chunks = []
            for chunk in result.chunks:
                if not isinstance(chunk.op, cls):
                    # some chunks are merged, get the inputs
                    chunks.extend(chunk.inputs)
                else:
                    chunks.append(chunk)
            for chunk in chunks:
                chunk.op.priority = -chunk.index[0]
            n_chunk = len(chunks)
            ctx = get_context()
            if ctx:
                name = str(uuid.uuid4())
                ctx.create_remote_object(name, _IncrementalIndexRecorder, n_chunk)
                for chunk in chunks:
                    chunk.op.incremental_index_recorder_name = name

    @classmethod
    def pre_execute(cls, ctx: Union[dict, Context], op: OperandType):
        out = op.outputs[0]
        if (
            op.incremental_index
            and isinstance(out.index_value.value, IndexValue.RangeIndex)
            and getattr(op, "incremental_index_recorder_name", None)
        ):
            index = out.index[0]
            recorder_name = op.incremental_index_recorder_name
            recorder = ctx.get_remote_object(recorder_name)
            recorder.add_waiter(index)

    @classmethod
    def post_execute(cls, ctx: Union[dict, Context], op: OperandType):
        out = op.outputs[0]
        result = ctx[out.key]
        if (
            op.incremental_index
            and isinstance(out.index_value.value, IndexValue.RangeIndex)
            and getattr(op, "incremental_index_recorder_name", None)
        ):
            recorder_name = op.incremental_index_recorder_name
            recorder = ctx.get_remote_object(recorder_name)
            index = out.index[0]
            recorder.finish(index, len(result))
            # wait for previous chunks to finish, then update index
            size, can_destroy = recorder.wait(index)
            result.index += size
            if can_destroy:
                try:
                    ctx.destroy_remote_object(recorder_name)
                except ActorNotExist:
                    pass


def merge_small_files(
    df: TileableType,
    n_sample_file: int = 10,
    merged_file_size: Union[int, float, str] = None,
) -> TileableType:
    from ..merge import DataFrameConcat

    if len(df.chunks) < n_sample_file:
        # if number of chunks is small(less than `n_sample_file`,
        # skip this process
        return df

    if merged_file_size is not None:
        merged_file_size = parse_readable_size(merged_file_size)[0]
    else:
        # Estimated size is relatively large than the real one,
        # so we double the merged size
        merged_file_size = options.chunk_store_limit * 2
    # sample files whose size equals `n_sample_file`
    sampled_chunks = np.random.choice(df.chunks, n_sample_file)
    max_chunk_size = 0
    ctx = dict()
    for sampled_chunk in sampled_chunks:
        sampled_chunk.op.estimate_size(ctx, sampled_chunk.op)
        size = ctx[sampled_chunk.key][0]
        max_chunk_size = max(max_chunk_size, size)
    to_merge_size = merged_file_size // max_chunk_size
    if to_merge_size < 2:
        return df
    # merge files
    n_new_chunks = np.ceil(len(df.chunks) / to_merge_size)
    new_chunks = []
    new_nsplit = []
    for i, chunks in enumerate(np.array_split(df.chunks, n_new_chunks)):
        chunk_size = sum(c.shape[0] for c in chunks)
        kw = dict(
            dtypes=chunks[0].dtypes,
            index_value=merge_index_value({c.index: c.index_value for c in chunks}),
            columns_value=chunks[0].columns_value,
            shape=(chunk_size, chunks[0].shape[1]),
            index=(i, 0),
        )
        new_chunk = DataFrameConcat(output_types=[OutputType.dataframe]).new_chunk(
            chunks.tolist(), **kw
        )
        new_chunks.append(new_chunk)
        new_nsplit.append(chunk_size)
    new_op = df.op.copy()
    params = df.params.copy()
    params["chunks"] = new_chunks
    params["nsplits"] = (tuple(new_nsplit), df.nsplits[1])
    return new_op.new_dataframe(df.op.inputs, kws=[params])



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from ...resource import cpu_count, cuda_count, mem_total, Resource
from ...services import NodeRole
from ...utils import get_next_port
from .cmdline import OscarCommandRunner
from .local import start_worker, stop_worker
from .pool import create_worker_actor_pool


class WorkerCommandRunner(OscarCommandRunner):
    command_description = "Mars Worker"
    node_role = NodeRole.WORKER

    def __init__(self):
        super().__init__()
        self.band_to_resource = dict()
        self.cuda_devices = []
        self.n_io_process = 1

    def config_args(self, parser):
        super().config_args(parser)
        parser.add_argument("--n-cpu", help="num of CPU to use", default="auto")
        parser.add_argument(
            "--mem-bytes", help="bytes of memory to use", default="auto"
        )
        parser.add_argument("--n-io-process", help="num of IO processes", default="1")
        parser.add_argument(
            "--cuda-devices",
            help="CUDA device to use, if not specified, will use "
            "all available devices",
            default="auto",
        )

    def parse_args(self, parser, argv, environ=None):
        environ = environ or os.environ
        args = super().parse_args(parser, argv, environ=environ)

        if (
            self.config.get("cluster", {}).get("backend", "fixed") == "fixed"
            and not args.supervisors
        ):  # pragma: no cover
            raise ValueError("--supervisors is needed to start Mars Worker")

        if args.endpoint is None:
            args.endpoint = f"{args.host}:{get_next_port()}"
        self.n_io_process = int(args.n_io_process)

        n_cpu = cpu_count() if args.n_cpu == "auto" else int(args.n_cpu)
        mem_bytes = mem_total() if args.mem_bytes == "auto" else int(args.mem_bytes)

        if "CUDA_VISIBLE_DEVICES" in os.environ:  # pragma: no cover
            args.cuda_devices = os.environ["CUDA_VISIBLE_DEVICES"].strip()

        if args.cuda_devices == "auto":
            self.cuda_devices = list(range(cuda_count()))
        elif args.cuda_devices.strip() == "":  # pragma: no cover
            # allow using CPU only
            self.cuda_devices = []
        else:  # pragma: no cover
            self.cuda_devices = [int(i) for i in args.cuda_devices.split(",")]

        self.band_to_resource = band_to_resource = dict()
        band_to_resource["numa-0"] = Resource(num_cpus=n_cpu, mem_bytes=mem_bytes)
        for i in self.cuda_devices:  # pragma: no cover
            band_to_resource[f"gpu-{i}"] = Resource(num_gpus=1)

        storage_config = self.config["storage"] = self.config.get("storage", {})
        backends = storage_config["backends"] = storage_config.get("backends", [])
        plasma_config = storage_config["plasma"] = storage_config.get("plasma", {})
        disk_config = storage_config["disk"] = storage_config.get("disk", {})
        if "MARS_CACHE_MEM_SIZE" in environ:
            plasma_config["store_memory"] = environ["MARS_CACHE_MEM_SIZE"]
        if "MARS_PLASMA_DIRS" in environ:
            plasma_config["plasma_directory"] = environ["MARS_PLASMA_DIRS"]
        if "MARS_SPILL_DIRS" in environ:
            backends.append("disk")
            disk_config["root_dirs"] = environ["MARS_SPILL_DIRS"]

        return args

    async def create_actor_pool(self):
        return await create_worker_actor_pool(
            self.args.endpoint,
            self.band_to_resource,
            ports=self.ports,
            n_io_process=self.n_io_process,
            modules=list(self.args.load_modules),
            logging_conf=self.logging_conf,
            cuda_devices=self.cuda_devices,
            subprocess_start_method="forkserver" if os.name != "nt" else "spawn",
            metrics=self.config.get("metrics", {}),
        )

    async def start_services(self):
        return await start_worker(
            self.pool.external_address,
            self.args.supervisors,
            self.band_to_resource,
            list(self.args.load_modules),
            self.config,
        )

    async def stop_services(self):
        return await stop_worker(self.pool.external_address, self.config)


main = WorkerCommandRunner()

if __name__ == "__main__":  # pragma: no branch
    main()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np
import pandas as pd

from ... import opcodes
from ...config import options
from ...core import ENTITY_TYPE, Entity, OutputType, get_output_types
from ...core.operand import OperandStage
from ...serialization.serializables import StringField, AnyField, BoolField, Int64Field
from ..align import (
    align_dataframe_dataframe,
    align_dataframe_series,
    align_series_series,
)
from ..core import DATAFRAME_TYPE, SERIES_TYPE
from ..operands import DataFrameOperandMixin, DataFrameOperand
from ..utils import validate_axis


class FillNA(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.FILL_NA

    _value = AnyField(
        "value", on_serialize=lambda x: x.data if isinstance(x, Entity) else x
    )
    _method = StringField("method")
    _axis = AnyField("axis")
    _limit = Int64Field("limit")
    _downcast = AnyField("downcast")
    _use_inf_as_na = BoolField("use_inf_as_na")

    _output_limit = Int64Field("output_limit")

    def __init__(
        self,
        value=None,
        method=None,
        axis=None,
        limit=None,
        downcast=None,
        use_inf_as_na=None,
        output_types=None,
        output_limit=None,
        **kw
    ):
        super().__init__(
            _value=value,
            _method=method,
            _axis=axis,
            _limit=limit,
            _downcast=downcast,
            _use_inf_as_na=use_inf_as_na,
            _output_types=output_types,
            _output_limit=output_limit,
            **kw
        )

    @property
    def value(self):
        return self._value

    @property
    def method(self):
        return self._method

    @property
    def axis(self):
        return self._axis

    @property
    def limit(self):
        return self._limit

    @property
    def downcast(self):
        return self._downcast

    @property
    def use_inf_as_na(self):
        return self._use_inf_as_na

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self._method is None and len(inputs) > 1:
            self._value = self._inputs[1]

    @property
    def output_limit(self):
        return self._output_limit or 1

    @staticmethod
    def _get_first_slice(op, df, end):
        if op.method == "bfill":
            if op.output_types[0] == OutputType.series:
                return df.iloc[:end]
            else:
                if op.axis == 1:
                    return df.iloc[:, :end]
                else:
                    return df.iloc[:end, :]
        else:
            if op.output_types[0] == OutputType.series:
                return df.iloc[-end:]
            else:
                if op.axis == 1:
                    return df.iloc[:, -end:]
                else:
                    return df.iloc[-end:, :]

    @classmethod
    def _execute_map(cls, ctx, op):
        input_data = ctx[op.inputs[0].key]
        limit = op.limit
        axis = op.axis
        method = op.method

        filled = input_data.fillna(
            method=method, axis=axis, limit=limit, downcast=op.downcast
        )
        ctx[op.outputs[0].key] = cls._get_first_slice(op, filled, 1)
        del filled

    @classmethod
    def _execute_combine(cls, ctx, op):
        axis = op.axis
        method = op.method
        limit = op.limit

        input_data = ctx[op.inputs[0].key]
        if limit is not None:
            n_summaries = (len(op.inputs) - 1) // 2
            summaries = [ctx[inp.key] for inp in op.inputs[1 : 1 + n_summaries]]
        else:
            summaries = [ctx[inp.key] for inp in op.inputs[1:]]

        if not summaries:
            ctx[op.outputs[0].key] = input_data.fillna(
                method=method, axis=axis, limit=limit, downcast=op.downcast
            )
            return

        valid_summary = cls._get_first_slice(
            op, pd.concat(summaries, axis=axis).fillna(method=method, axis=axis), 1
        )

        if method == "bfill":
            concat_df = pd.concat([input_data, valid_summary], axis=axis)
        else:
            concat_df = pd.concat([valid_summary, input_data], axis=axis)

        concat_df.fillna(
            method=method, axis=axis, inplace=True, limit=limit, downcast=op.downcast
        )
        ctx[op.outputs[0].key] = cls._get_first_slice(op, concat_df, -1)

    @classmethod
    def execute(cls, ctx, op):
        try:
            pd.set_option("mode.use_inf_as_na", op.use_inf_as_na)
            if op.stage == OperandStage.map:
                cls._execute_map(ctx, op)
            elif op.stage == OperandStage.combine:
                cls._execute_combine(ctx, op)
            else:
                input_data = ctx[op.inputs[0].key]
                value = getattr(op, "value", None)
                if isinstance(op.value, ENTITY_TYPE):
                    value = ctx[op.value.key]
                if not isinstance(input_data, pd.Index):
                    ctx[op.outputs[0].key] = input_data.fillna(
                        value=value,
                        method=op.method,
                        axis=op.axis,
                        limit=op.limit,
                        downcast=op.downcast,
                    )
                else:
                    ctx[op.outputs[0].key] = input_data.fillna(
                        value=value, downcast=op.downcast
                    )
        finally:
            pd.reset_option("mode.use_inf_as_na")

    @classmethod
    def _tile_one_by_one(cls, op):
        in_df = op.inputs[0]
        in_value_df = op.value if isinstance(op.value, ENTITY_TYPE) else None
        df = op.outputs[0]

        new_chunks = []
        for c in in_df.chunks:
            inputs = [c] if in_value_df is None else [c, in_value_df.chunks[0]]
            kw = c.params
            new_op = op.copy().reset_key()
            new_chunks.append(new_op.new_chunk(inputs, **kw))

        kw = df.params.copy()
        kw.update(dict(chunks=new_chunks, nsplits=in_df.nsplits))
        new_op = op.copy().reset_key()
        return new_op.new_tileables(op.inputs, **kw)

    @classmethod
    def _build_combine(cls, op, input_chunks, summary_chunks, idx, is_forward=True):
        c = input_chunks[idx]

        summaries_to_concat = []

        idx_range = list(
            range(idx) if is_forward else range(idx + 1, len(summary_chunks))
        )
        for i in idx_range:
            summaries_to_concat.append(summary_chunks[i])

        new_chunk_op = op.copy().reset_key()
        new_chunk_op.stage = OperandStage.combine

        chunks_to_concat = [c] + summaries_to_concat
        return new_chunk_op.new_chunk(chunks_to_concat, **c.params)

    @classmethod
    def _tile_directional_dataframe(cls, op):
        in_df = op.inputs[0]
        df = op.outputs[0]
        is_forward = op.method == "ffill"

        n_rows, n_cols = in_df.chunk_shape

        # map to get individual results and summaries
        src_chunks = np.empty(in_df.chunk_shape, dtype=object)
        summary_chunks = np.empty(in_df.chunk_shape, dtype=object)
        for c in in_df.chunks:
            new_chunk_op = op.copy().reset_key()
            new_chunk_op.stage = OperandStage.map
            if op.axis == 1:
                summary_shape = (c.shape[0], 1)
            else:
                summary_shape = (1, c.shape[1])
            src_chunks[c.index] = c
            summary_chunks[c.index] = new_chunk_op.new_chunk(
                [c], shape=summary_shape, dtypes=df.dtypes
            )

        # combine summaries into results
        output_chunk_array = np.empty(in_df.chunk_shape, dtype=object)
        if op.axis == 1:
            for row in range(n_rows):
                row_src = src_chunks[row, :]
                row_summaries = summary_chunks[row, :]
                for col in range(n_cols):
                    output_chunk_array[row, col] = cls._build_combine(
                        op, row_src, row_summaries, col, is_forward
                    )
        else:
            for col in range(n_cols):
                col_src = src_chunks[:, col]
                col_summaries = summary_chunks[:, col]
                for row in range(n_rows):
                    output_chunk_array[row, col] = cls._build_combine(
                        op, col_src, col_summaries, row, is_forward
                    )

        output_chunks = list(output_chunk_array.reshape((n_rows * n_cols,)))
        new_op = op.copy().reset_key()
        return new_op.new_tileables(
            op.inputs,
            shape=in_df.shape,
            nsplits=in_df.nsplits,
            chunks=output_chunks,
            dtypes=df.dtypes,
            index_value=df.index_value,
            columns_value=df.columns_value,
        )

    @classmethod
    def _tile_directional_series(cls, op):
        in_series = op.inputs[0]
        series = op.outputs[0]
        forward = op.method == "ffill"

        # map to get individual results and summaries
        summary_chunks = np.empty(in_series.chunk_shape, dtype=object)
        for c in in_series.chunks:
            new_chunk_op = op.copy().reset_key()
            new_chunk_op.stage = OperandStage.map
            summary_chunks[c.index] = new_chunk_op.new_chunk(
                [c], shape=(1,), dtype=series.dtype
            )

        # combine summaries into results
        output_chunks = [
            cls._build_combine(op, in_series.chunks, summary_chunks, i, forward)
            for i in range(len(in_series.chunks))
        ]
        new_op = op.copy().reset_key()
        return new_op.new_tileables(
            op.inputs,
            shape=in_series.shape,
            nsplits=in_series.nsplits,
            chunks=output_chunks,
            dtype=series.dtype,
            index_value=series.index_value,
        )

    @classmethod
    def _tile_both_dataframes(cls, op):
        in_df = op.inputs[0]
        in_value = op.inputs[1]
        df = op.outputs[0]

        nsplits, out_shape, left_chunks, right_chunks = align_dataframe_dataframe(
            in_df, in_value
        )
        out_chunk_indexes = itertools.product(*(range(s) for s in out_shape))

        out_chunks = []
        for idx, left_chunk, right_chunk in zip(
            out_chunk_indexes, left_chunks, right_chunks
        ):
            out_chunk = (
                op.copy()
                .reset_key()
                .new_chunk([left_chunk, right_chunk], shape=(np.nan, np.nan), index=idx)
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_dataframes(
            op.inputs,
            df.shape,
            nsplits=tuple(tuple(ns) for ns in nsplits),
            chunks=out_chunks,
            dtypes=df.dtypes,
            index_value=df.index_value,
            columns_value=df.columns_value,
        )

    @classmethod
    def _tile_dataframe_series(cls, op):
        left, right = op.inputs[0], op.inputs[1]
        df = op.outputs[0]

        nsplits, out_shape, left_chunks, right_chunks = align_dataframe_series(
            left, right, axis=1
        )
        out_chunk_indexes = itertools.product(*(range(s) for s in out_shape))

        out_chunks = []
        for out_idx, df_chunk in zip(out_chunk_indexes, left_chunks):
            series_chunk = right_chunks[out_idx[1]]
            kw = dict(
                shape=(nsplits[0][out_idx[0]], nsplits[1][out_idx[1]]),
                index_value=df_chunk.index_value,
                columns_value=df_chunk.columns_value,
            )
            out_chunk = (
                op.copy()
                .reset_key()
                .new_chunk([df_chunk, series_chunk], index=out_idx, **kw)
            )
            out_chunks.append(out_chunk)

        new_op = op.copy().reset_key()
        return new_op.new_dataframes(
            op.inputs,
            df.shape,
            nsplits=tuple(tuple(ns) for ns in nsplits),
            chunks=out_chunks,
            dtypes=df.dtypes,
            index_value=df.index_value,
            columns_value=df.columns_value,
        )

    @classmethod
    def _tile_both_series(cls, op):
        left, right = op.inputs[0], op.inputs[1]
        df = op.outputs[0]

        nsplits, out_shape, left_chunks, right_chunks = align_series_series(left, right)

        out_chunks = []
        for idx, left_chunk, right_chunk in zip(
            range(out_shape[0]), left_chunks, right_chunks
        ):
            out_chunk = (
                op.copy()
                .reset_key()
                .new_chunk(
                    [left_chunk, right_chunk],
                    index_value=left_chunk.index_value,
                    shape=(np.nan,),
                    index=(idx,),
                )
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_seriess(
            op.inputs,
            df.shape,
            nsplits=tuple(tuple(ns) for ns in nsplits),
            chunks=out_chunks,
            dtype=df.dtype,
            index_value=df.index_value,
            name=df.name,
        )

    @classmethod
    def tile(cls, op):
        in_df = op.inputs[0]
        if len(in_df.chunks) == 1 and (
            not isinstance(op.value, ENTITY_TYPE) or len(op.value.chunks) == 1
        ):
            return cls._tile_one_by_one(op)
        elif op.method is not None:
            if op.output_types[0] == OutputType.dataframe:
                return cls._tile_directional_dataframe(op)
            else:
                return cls._tile_directional_series(op)
        elif not isinstance(op.value, ENTITY_TYPE):
            return cls._tile_one_by_one(op)
        elif isinstance(op.value, DATAFRAME_TYPE):
            return cls._tile_both_dataframes(op)
        elif op.output_types[0] == OutputType.dataframe:
            return cls._tile_dataframe_series(op)
        else:
            return cls._tile_both_series(op)

    def __call__(self, a, value_df=None):
        method = getattr(self, "method", None)
        if method == "backfill":
            method = "bfill"
        elif method == "pad":
            method = "ffill"
        self._method = method
        axis = getattr(self, "axis", None) or 0
        self._axis = validate_axis(axis, a)

        inputs = [a]
        if value_df is not None:
            inputs.append(value_df)
        if isinstance(a, DATAFRAME_TYPE):
            return self.new_dataframe(
                inputs,
                shape=a.shape,
                dtypes=a.dtypes,
                index_value=a.index_value,
                columns_value=a.columns_value,
            )
        elif isinstance(a, SERIES_TYPE):
            return self.new_series(
                inputs,
                shape=a.shape,
                dtype=a.dtype,
                index_value=a.index_value,
                name=a.name,
            )
        else:
            return self.new_index(
                inputs,
                shape=a.shape,
                dtype=a.dtype,
                index_value=a.index_value,
                name=a.name,
                names=a.names,
            )


def fillna(
    df, value=None, method=None, axis=None, inplace=False, limit=None, downcast=None
):
    """
    Fill NA/NaN values using the specified method.

    Parameters
    ----------
    value : scalar, dict, Series, or DataFrame
        Value to use to fill holes (e.g. 0), alternately a
        dict/Series/DataFrame of values specifying which value to use for
        each index (for a Series) or column (for a DataFrame).  Values not
        in the dict/Series/DataFrame will not be filled. This value cannot
        be a list.
    method : {'backfill', 'bfill', 'pad', 'ffill', None}, default None
        Method to use for filling holes in reindexed Series
        pad / ffill: propagate last valid observation forward to next valid
        backfill / bfill: use next valid observation to fill gap.
    axis : {0 or 'index', 1 or 'columns'}
        Axis along which to fill missing values.
    inplace : bool, default False
        If True, fill in-place. Note: this will modify any
        other views on this object (e.g., a no-copy slice for a column in a
        DataFrame).
    limit : int, default None
        If method is specified, this is the maximum number of consecutive
        NaN values to forward/backward fill. In other words, if there is
        a gap with more than this number of consecutive NaNs, it will only
        be partially filled. If method is not specified, this is the
        maximum number of entries along the entire axis where NaNs will be
        filled. Must be greater than 0 if not None.
    downcast : dict, default is None
        A dict of item->dtype of what to downcast if possible,
        or the string 'infer' which will try to downcast to an appropriate
        equal type (e.g. float64 to int64 if possible).

    Returns
    -------
    DataFrame or None
        Object with missing values filled or None if ``inplace=True``.

    See Also
    --------
    interpolate : Fill NaN values using interpolation.
    reindex : Conform object to new index.
    asfreq : Convert TimeSeries to specified frequency.

    Examples
    --------
    >>> import mars.tensor as mt
            >>> import mars.dataframe as md
    >>> df = md.DataFrame([[mt.nan, 2, mt.nan, 0],
    ...                    [3, 4, mt.nan, 1],
    ...                    [mt.nan, mt.nan, mt.nan, 5],
    ...                    [mt.nan, 3, mt.nan, 4]],
    ...                   columns=list('ABCD'))
    >>> df.execute()
         A    B   C  D
    0  NaN  2.0 NaN  0
    1  3.0  4.0 NaN  1
    2  NaN  NaN NaN  5
    3  NaN  3.0 NaN  4

    Replace all NaN elements with 0s.

    >>> df.fillna(0).execute()
        A   B   C   D
    0   0.0 2.0 0.0 0
    1   3.0 4.0 0.0 1
    2   0.0 0.0 0.0 5
    3   0.0 3.0 0.0 4

    We can also propagate non-null values forward or backward.

    >>> df.fillna(method='ffill').execute()
        A   B   C   D
    0   NaN 2.0 NaN 0
    1   3.0 4.0 NaN 1
    2   3.0 4.0 NaN 5
    3   3.0 3.0 NaN 4

    Replace all NaN elements in column 'A', 'B', 'C', and 'D', with 0, 1,
    2, and 3 respectively.

    >>> values = {'A': 0, 'B': 1, 'C': 2, 'D': 3}
    >>> df.fillna(value=values).execute()
        A   B   C   D
    0   0.0 2.0 2.0 0
    1   3.0 4.0 2.0 1
    2   0.0 1.0 2.0 5
    3   0.0 3.0 2.0 4
    """
    if value is None and method is None:
        raise ValueError("Must specify a fill 'value' or 'method'.")
    elif value is not None and method is not None:
        raise ValueError("Cannot specify both 'value' and 'method'.")

    if isinstance(df, SERIES_TYPE) and isinstance(
        value, (DATAFRAME_TYPE, pd.DataFrame)
    ):
        raise ValueError(
            '"value" parameter must be a scalar, dict or Series, but you passed a "%s"'
            % type(value).__name__
        )

    if downcast is not None:
        raise NotImplementedError(
            'Currently argument "downcast" is not implemented yet'
        )
    if limit is not None:
        raise NotImplementedError('Currently argument "limit" is not implemented yet')

    if isinstance(value, ENTITY_TYPE):
        value, value_df = None, value
    else:
        value_df = None

    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = FillNA(
        value=value,
        method=method,
        axis=axis,
        limit=limit,
        downcast=downcast,
        use_inf_as_na=use_inf_as_na,
        output_types=get_output_types(df),
    )
    out_df = op(df, value_df=value_df)
    if inplace:
        df.data = out_df.data
    else:
        return out_df


def ffill(df, axis=None, inplace=False, limit=None, downcast=None):
    """
    Synonym for :meth:`DataFrame.fillna` with ``method='ffill'``.

    Returns
    -------
    {klass} or None
        Object with missing values filled or None if ``inplace=True``.
    """
    return fillna(
        df, method="ffill", axis=axis, inplace=inplace, limit=limit, downcast=downcast
    )


def bfill(df, axis=None, inplace=False, limit=None, downcast=None):
    """
    Synonym for :meth:`DataFrame.fillna` with ``method='bfill'``.

    Returns
    -------
    {klass} or None
        Object with missing values filled or None if ``inplace=True``.
    """
    return fillna(
        df, method="bfill", axis=axis, inplace=inplace, limit=limit, downcast=downcast
    )


def index_fillna(index, value=None, downcast=None):
    """
    Fill NA/NaN values with the specified value.

    Parameters
    ----------
    value : scalar
        Scalar value to use to fill holes (e.g. 0).
        This value cannot be a list-likes.
    downcast : dict, default is None
        A dict of item->dtype of what to downcast if possible,
        or the string 'infer' which will try to downcast to an appropriate
        equal type (e.g. float64 to int64 if possible).

    Returns
    -------
    Index

    See Also
    --------
    DataFrame.fillna : Fill NaN values of a DataFrame.
    Series.fillna : Fill NaN Values of a Series.
    """
    if isinstance(value, (list, pd.Series, SERIES_TYPE)):
        raise ValueError("'value' must be a scalar, passed: %s" % type(value))

    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = FillNA(
        value=value,
        downcast=downcast,
        use_inf_as_na=use_inf_as_na,
        output_types=get_output_types(index),
    )
    return op(index)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ..datasource import tensor as astensor
from ..base.broadcast_to import broadcast_to
from ..base.swapaxes import swapaxes


def average(a, axis=None, weights=None, returned=False):
    """
    Compute the weighted average along the specified axis.

    Parameters
    ----------
    a : array_like
        Tensor containing data to be averaged. If `a` is not a tensor, a
        conversion is attempted.
    axis : None or int or tuple of ints, optional
        Axis or axes along which to average `a`.  The default,
        axis=None, will average over all of the elements of the input tensor.
        If axis is negative it counts from the last to the first axis.

        If axis is a tuple of ints, averaging is performed on all of the axes
        specified in the tuple instead of a single axis or all the axes as
        before.
    weights : array_like, optional
        A tensor of weights associated with the values in `a`. Each value in
        `a` contributes to the average according to its associated weight.
        The weights tensor can either be 1-D (in which case its length must be
        the size of `a` along the given axis) or of the same shape as `a`.
        If `weights=None`, then all data in `a` are assumed to have a
        weight equal to one.
    returned : bool, optional
        Default is `False`. If `True`, the tuple (`average`, `sum_of_weights`)
        is returned, otherwise only the average is returned.
        If `weights=None`, `sum_of_weights` is equivalent to the number of
        elements over which the average is taken.


    Returns
    -------
    average, [sum_of_weights] : tensor_type or double
        Return the average along the specified axis. When returned is `True`,
        return a tuple with the average as the first element and the sum
        of the weights as the second element. The return type is `Float`
        if `a` is of integer type, otherwise it is of the same type as `a`.
        `sum_of_weights` is of the same type as `average`.

    Raises
    ------
    ZeroDivisionError
        When all weights along axis are zero. See `numpy.ma.average` for a
        version robust to this type of error.
    TypeError
        When the length of 1D `weights` is not the same as the shape of `a`
        along axis.

    See Also
    --------
    mean

    Examples
    --------
    >>> import mars.tensor as mt

    >>> data = list(range(1,5))
    >>> data
    [1, 2, 3, 4]
    >>> mt.average(data).execute()
    2.5
    >>> mt.average(range(1,11), weights=range(10,0,-1)).execute()
    4.0

    >>> data = mt.arange(6).reshape((3,2))
    >>> data.execute()
    array([[0, 1],
           [2, 3],
           [4, 5]])
    >>> mt.average(data, axis=1, weights=[1./4, 3./4]).execute()
    array([ 0.75,  2.75,  4.75])
    >>> mt.average(data, weights=[1./4, 3./4]).execute()
    Traceback (most recent call last):
    ...
    TypeError: Axis must be specified when shapes of a and weights differ.

    """
    from ..arithmetic import truediv, multiply

    a = astensor(a)

    if weights is None:
        avg = a.mean(axis)
        scl = avg.dtype.type(a.size / avg.size)
    else:
        wgt = astensor(weights)

        if issubclass(a.dtype.type, (np.integer, np.bool_)):
            result_dtype = np.result_type(a.dtype, wgt.dtype, "f8")
        else:
            result_dtype = np.result_type(a.dtype, wgt.dtype)

        # sanity checks
        if a.shape != wgt.shape:
            if axis is None:
                raise TypeError(
                    "Axis must be specified when shapes of a and weights differ."
                )
            if wgt.ndim != 1:
                raise TypeError(
                    "1D weights expected when shapes of a and weights differ."
                )
            if wgt.shape[0] != a.shape[axis]:
                raise ValueError(
                    "Length of weights not compatible with specified axis."
                )

            # setup wgt to broadcast along axis
            wgt = broadcast_to(wgt, (a.ndim - 1) * (1,) + wgt.shape)
            wgt = swapaxes(wgt, -1, axis)

        scl = wgt.sum(axis=axis, dtype=result_dtype)
        with np.errstate(divide="raise"):
            avg = truediv(multiply(a, wgt, dtype=result_dtype).sum(axis), scl)

    if returned:
        if scl.shape != avg.shape:
            scl = broadcast_to(scl, avg.shape)
        return avg, scl
    else:
        return avg



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dask import is_dask_collection, optimize
from dask.bag import Bag

from .scheduler import mars_dask_get
from .utils import reduce
from ...remote import spawn


def convert_dask_collection(dc):
    """
    Convert dask collection object into mars.core.Object via remote API

    Parameters
    ----------
    dc: dask collection
        Dask collection object to be converted.

    Returns
    -------
    Object
        Mars Object.
    """
    if not is_dask_collection(dc):
        raise TypeError(f"'{type(dc).__name__}' object is not a valid dask collection")

    dc.__dask_graph__().validate()
    dsk = optimize(dc)[0].__dask_graph__()

    first_key = next(iter(dsk.keys()))
    if isinstance(first_key, str):
        key = [first_key]
    elif isinstance(first_key, tuple):
        key = sorted(
            [i for i in dsk.keys() if i[0] == first_key[0]], key=lambda x: x[1]
        )
    else:
        raise ValueError(
            f"Dask collection object seems be broken, with unexpected key type:'{type(first_key).__name__}'"
        )
    res = reduce(mars_dask_get(dsk, [key]))
    if isinstance(dc, Bag):
        return spawn(lambda x: list(x[0][0]), args=(res,))
    else:
        return res



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import weakref
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Optional, Tuple, Type

from ...core import OperandType, EntityType, enter_mode
from ...core.graph import EntityGraph
from ...core.operand import Operand


class OptimizationRecordType(Enum):
    replace = 0
    new = 1
    delete = 2


@dataclass
class OptimizationRecord:
    original_entity: EntityType = None
    new_entity: EntityType = None
    record_type: OptimizationRecordType = None


class OptimizationRecords:
    _records: List[OptimizationRecord]
    _original_entity_to_records: Dict[EntityType, OptimizationRecord]

    def __init__(self):
        self._records = list()
        self._original_entity_to_records = dict()
        self._optimized_entity_to_records = dict()

    def append_record(self, record: OptimizationRecord):
        self._records.append(record)
        if record.record_type in (
            OptimizationRecordType.replace,
            OptimizationRecordType.delete,
        ):
            self._original_entity_to_records[record.original_entity] = record
        if record.record_type in (
            OptimizationRecordType.new,
            OptimizationRecordType.replace,
        ):
            self._optimized_entity_to_records[record.new_entity] = record

    def get_optimization_result(
        self, original_entity: EntityType, default: Optional[EntityType] = None
    ) -> EntityType:
        entity = original_entity
        if entity not in self._original_entity_to_records:
            return default
        while entity in self._original_entity_to_records:
            record = self._original_entity_to_records[entity]
            if record.record_type == OptimizationRecordType.replace:
                entity = record.new_entity
            else:
                assert record.record_type == OptimizationRecordType.delete
                return None
        return entity

    def get_original_entity(
        self, optimized_entity: EntityType, default: Optional[EntityType] = None
    ) -> EntityType:
        entity = optimized_entity
        if entity not in self._optimized_entity_to_records:
            return default
        while entity in self._optimized_entity_to_records:
            record = self._optimized_entity_to_records[entity]
            if record.record_type == OptimizationRecordType.replace:
                entity = record.original_entity
            else:
                assert record.record_type == OptimizationRecordType.new
                return None
        return entity


class OptimizationRule(ABC):
    _instances: Dict[
        Tuple[Type["OptimizationRule"], EntityGraph, OptimizationRecords],
        "OptimizationRule",
    ] = dict()
    _preds_to_remove = weakref.WeakKeyDictionary()

    def __init__(
        self,
        graph: EntityGraph,
        records: OptimizationRecords,
        optimizer_cls: Type["Optimizer"],
    ):
        self._graph = graph
        self._records = records
        self._optimizer_cls = optimizer_cls

    def __new__(
        cls,
        graph: EntityGraph,
        records: OptimizationRecords,
        optimizer_cls: Type["Optimizer"],
    ):
        if (cls, graph, records) in cls._instances:
            return cls._instances[cls, graph, records]
        inst = cls._instances[cls, graph, records] = object.__new__(cls)
        return inst

    @abstractmethod
    def match(self, op: OperandType) -> bool:
        """
        If this operand matches this rule.

        Parameters
        ----------
        op : OperandType
            Operand.

        Returns
        -------
        matched : bool
            Matched rule or not.
        """

    @abstractmethod
    def apply(self, op: OperandType):
        """
        Apply rule to an operand.

        Parameters
        ----------
        op : OperandType
            Operand
        """

    def _replace_node(self, original_node: EntityType, new_node: EntityType):
        predecessors = self._graph.predecessors(original_node)
        successors = self._graph.successors(original_node)
        self._graph.remove_node(original_node)
        self._graph.add_node(new_node)
        for pred in predecessors:
            self._graph.add_edge(pred, new_node)
        for succ in successors:
            self._graph.add_edge(new_node, succ)

    def _add_collapsable_predecessor(self, node: EntityType, predecessor: EntityType):
        pred_original = self._records.get_original_entity(predecessor, predecessor)
        if predecessor not in self._preds_to_remove:
            self._preds_to_remove[pred_original] = {node}
        else:
            self._preds_to_remove[pred_original].add(node)

    def _remove_collapsable_predecessors(self, node: EntityType):
        node = self._records.get_optimization_result(node) or node
        preds_opt_to_remove = []
        for pred in self._graph.predecessors(node):
            pred_original = self._records.get_original_entity(pred, pred)
            pred_opt = self._records.get_optimization_result(pred, pred)

            if pred_opt in self._graph.results or pred_original in self._graph.results:
                continue
            affect_succ = self._preds_to_remove.get(pred_original) or []
            affect_succ_opt = [
                self._records.get_optimization_result(s, s) for s in affect_succ
            ]
            if all(s in affect_succ_opt for s in self._graph.successors(pred)):
                preds_opt_to_remove.append((pred_original, pred_opt))

        for pred_original, pred_opt in preds_opt_to_remove:
            self._graph.remove_node(pred_opt)
            self._records.append_record(
                OptimizationRecord(pred_original, None, OptimizationRecordType.delete)
            )


class Optimizer(ABC):
    _rules: List[Type[OptimizationRule]]
    _op_to_rules: Dict[Type[OperandType], List[Type[OptimizationRule]]]

    @classmethod
    def register_rule(
        cls, operand_types: List[Type[OperandType]], rule: Type[OptimizationRule]
    ):
        if not hasattr(cls, "_rules"):
            cls._rules = []
        cls._rules.append(rule)

        if not hasattr(cls, "_op_to_rules"):
            cls._op_to_rules = defaultdict(list)
        for operand_type in operand_types:
            cls._op_to_rules[operand_type].append(rule)

    @classmethod
    def get_rule_types(
        cls, operand_type: Type[OperandType]
    ) -> List[Type[OptimizationRule]]:
        rule_types = cls._op_to_rules.get(operand_type, None)
        if rule_types is None:
            for op_cls in operand_type.__mro__:
                if op_cls is Operand:
                    break
                rule_types = cls._op_to_rules.get(op_cls)
                if rule_types is not None:
                    break
            cls._op_to_rules[operand_type] = rule_types or []
        return rule_types

    @classmethod
    def _replace_inputs(cls, graph: EntityGraph, records: OptimizationRecords):
        for node in graph:
            for succ in graph.successors(node):
                input_optimized = False
                new_inputs = []
                for inp in succ.inputs:
                    optimized = records.get_optimization_result(inp)
                    if optimized is None:
                        optimized = inp
                    if optimized is not inp:
                        input_optimized = True
                    new_inputs.append(optimized)
                if input_optimized:
                    succ.inputs = new_inputs

    @classmethod
    @enter_mode(build=True)
    def optimize(cls, graph: EntityGraph) -> OptimizationRecords:
        """
        Optimize a graph.

        Parameters
        ----------
        graph : EntityGraph
            Tileable or chunk graph.

        Returns
        -------
        optimization_records : OptimizationRecords
            Optimization records.
        """
        records = OptimizationRecords()
        optimized = False
        for rule_type in cls._rules:
            visited = set()
            for entity in list(graph.topological_iter()):
                op = entity.op
                if op in visited:
                    continue
                visited.add(op)

                rule_types = cls.get_rule_types(type(op)) or []
                if rule_type not in rule_types:
                    continue

                rule = rule_type(graph, records, cls)
                if entity not in graph:  # pragma: no cover
                    # maybe removed during optimization
                    continue
                if rule.match(op):
                    optimized = True
                    rule.apply(op)
        if optimized:
            cls._replace_inputs(graph, records)
            new_results = []
            for result in graph.results:
                new_results.append(
                    records.get_optimization_result(result, default=result)
                )
            graph.results = new_results
        return records



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
from functools import partial

import numpy as np
import pandas as pd

from .... import opcodes
from ....serialization.serializables import BoolField
from ..aggregation import BaseDataFrameExpandingAgg

_stage_info = namedtuple(
    "_stage_info",
    (
        "map_groups",
        "map_sources",
        "combine_sources",
        "combine_columns",
        "combine_funcs",
        "key_to_funcs",
        "valid_columns",
        "min_periods_func_name",
    ),
)

_cum_alpha_coeff_func = "_cum_alpha_coeff"
_cum_square_alpha_coeff_func = "_cum_square_alpha_coeff"


def _add_pred_results(pred_results, local_results, axis=0):
    if pred_results[0].ndim == 1:
        df_filler = 0
    else:
        df_filler = pred_results[0].iloc[-1, :].dropna()
        df_filler[:] = 0

    new_locals = []
    combine_axis = pred_results[0].ndim - axis - 1
    for pred_result, local_result in zip(pred_results, local_results):
        local_result = local_result.fillna(df_filler, axis=axis)
        new_locals.append(
            local_result.add(pred_result.sum(axis=axis), axis=combine_axis)
        )
    return new_locals


def _combine_arithmetic(pred_results, local_results, axis=0):
    if pred_results is None:
        return local_results[0]
    return _add_pred_results(pred_results, local_results, axis=axis)[0]


def _combine_minmax(pred_results, local_results, axis=0, fun_name=None):
    if pred_results is None:
        return local_results[0]

    pred_size = len(pred_results[0])
    con = pd.concat([pred_results[0], local_results[0]], axis=axis)
    result = con.expanding(axis=axis).agg(fun_name)
    if result.ndim == 2:
        return result.iloc[pred_size:, :] if axis == 0 else result.iloc[:, pred_size:]
    else:
        return result.iloc[pred_size:]


def _combine_mean(pred_results, local_results, axis=0):
    local_sum_data, local_count_data = local_results

    if pred_results is not None:
        local_sum_data, local_count_data = _add_pred_results(
            pred_results, local_results, axis=axis
        )
    return local_sum_data / local_count_data


def _combine_var(pred_results, local_results, axis=0):
    local_sum_data, local_count_data, local_var_data = local_results
    if pred_results is None:
        return local_var_data * local_count_data / (local_count_data - 1)

    pred_sum_data, pred_count_data, pred_var_data = pred_results

    local_sum_square = (
        local_count_data * local_var_data + local_sum_data**2 / local_count_data
    )
    pred_sum_square = (
        pred_count_data * pred_var_data + pred_sum_data**2 / pred_count_data
    )

    local_sum_square, local_sum_data, local_count_data = _add_pred_results(
        [pred_sum_square, pred_sum_data, pred_count_data],
        [local_sum_square, local_sum_data, local_count_data],
        axis=axis,
    )

    return (local_sum_square - local_sum_data**2 / local_count_data) / (
        local_count_data - 1
    )


def _combine_std(pred_results, local_results, axis=0):
    return np.sqrt(_combine_var(pred_results, local_results, axis=axis))


class DataFrameExpandingAgg(BaseDataFrameExpandingAgg):
    _op_type_ = opcodes.EXPANDING_AGG

    _center = BoolField("center")

    def __init__(self, center=None, **kw):
        super().__init__(_center=center, **kw)

    @property
    def center(self):
        return self._center

    @classmethod
    def _get_stage_functions(cls, op: "DataFrameExpandingAgg", func):
        if func == "_data_count":
            return ["count"], _combine_arithmetic
        elif func in ("sum", "prod", "count"):
            return [func], _combine_arithmetic
        elif func in ("min", "max"):
            return [func], partial(_combine_minmax, fun_name=func)
        elif func == "mean":
            return ["sum", "count"], _combine_mean
        elif func in {"var", "std"}:
            return (
                ["sum", "count", "var"],
                _combine_var if func == "var" else _combine_std,
            )
        else:  # pragma: no cover
            raise NotImplementedError

    @classmethod
    def _execute_map_function(cls, op: "DataFrameExpandingAgg", func, in_data):
        min_periods = 1 if op.min_periods > 0 else 0

        expanding = in_data.expanding(
            min_periods=min_periods, center=op.center, axis=op.axis
        )
        if func == "var":
            result = expanding.var(ddof=0)
        else:
            result = expanding.agg(func)

        if op.output_agg:
            summary = result.iloc[len(result) - 1 : len(result)]
        else:
            summary = None
        return result, summary

    @classmethod
    def _execute_combine_function(
        cls, op: "DataFrameExpandingAgg", func, pred_inputs, local_inputs, func_cols
    ):
        return func(pred_inputs, local_inputs, axis=op.axis)

    @classmethod
    def _execute_raw_function(cls, op: "DataFrameExpandingAgg", in_data):
        expanding = in_data.expanding(
            min_periods=op.min_periods, center=op.center, axis=op.axis
        )
        return expanding.agg(op.func)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..datasource import tensor as astensor
from .core import TensorComplexFFTNMixin, validate_fftn, TensorStandardFFTN


class TensorIFFTN(TensorStandardFFTN, TensorComplexFFTNMixin):
    _op_type_ = OperandDef.IFFTN

    def __init__(self, shape=None, axes=None, norm=None, **kw):
        super().__init__(_shape=shape, _axes=axes, _norm=norm, **kw)


def ifftn(a, s=None, axes=None, norm=None):
    """
    Compute the N-dimensional inverse discrete Fourier Transform.

    This function computes the inverse of the N-dimensional discrete
    Fourier Transform over any number of axes in an M-dimensional tensor by
    means of the Fast Fourier Transform (FFT).  In other words,
    ``ifftn(fftn(a)) == a`` to within numerical accuracy.
    For a description of the definitions and conventions used, see `mt.fft`.

    The input, analogously to `ifft`, should be ordered in the same way as is
    returned by `fftn`, i.e. it should have the term for zero frequency
    in all axes in the low-order corner, the positive frequency terms in the
    first half of all axes, the term for the Nyquist frequency in the middle
    of all axes and the negative frequency terms in the second half of all
    axes, in order of decreasingly negative frequency.

    Parameters
    ----------
    a : array_like
        Input tensor, can be complex.
    s : sequence of ints, optional
        Shape (length of each transformed axis) of the output
        (``s[0]`` refers to axis 0, ``s[1]`` to axis 1, etc.).
        This corresponds to ``n`` for ``ifft(x, n)``.
        Along any axis, if the given shape is smaller than that of the input,
        the input is cropped.  If it is larger, the input is padded with zeros.
        if `s` is not given, the shape of the input along the axes specified
        by `axes` is used.  See notes for issue on `ifft` zero padding.
    axes : sequence of ints, optional
        Axes over which to compute the IFFT.  If not given, the last ``len(s)``
        axes are used, or all axes if `s` is also not specified.
        Repeated indices in `axes` means that the inverse transform over that
        axis is performed multiple times.
    norm : {None, "ortho"}, optional
        Normalization mode (see `mt.fft`). Default is None.

    Returns
    -------
    out : complex Tensor
        The truncated or zero-padded input, transformed along the axes
        indicated by `axes`, or by a combination of `s` or `a`,
        as explained in the parameters section above.

    Raises
    ------
    ValueError
        If `s` and `axes` have different length.
    IndexError
        If an element of `axes` is larger than than the number of axes of `a`.

    See Also
    --------
    mt.fft : Overall view of discrete Fourier transforms, with definitions
         and conventions used.
    fftn : The forward *n*-dimensional FFT, of which `ifftn` is the inverse.
    ifft : The one-dimensional inverse FFT.
    ifft2 : The two-dimensional inverse FFT.
    ifftshift : Undoes `fftshift`, shifts zero-frequency terms to beginning
        of tensor.

    Notes
    -----
    See `mt.fft` for definitions and conventions used.

    Zero-padding, analogously with `ifft`, is performed by appending zeros to
    the input along the specified dimension.  Although this is the common
    approach, it might lead to surprising results.  If another form of zero
    padding is desired, it must be performed before `ifftn` is called.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.eye(4)
    >>> mt.fft.ifftn(mt.fft.fftn(a, axes=(0,)), axes=(1,)).execute()
    array([[ 1.+0.j,  0.+0.j,  0.+0.j,  0.+0.j],
           [ 0.+0.j,  1.+0.j,  0.+0.j,  0.+0.j],
           [ 0.+0.j,  0.+0.j,  1.+0.j,  0.+0.j],
           [ 0.+0.j,  0.+0.j,  0.+0.j,  1.+0.j]])


    Create and plot an image with band-limited frequency content:

    >>> import matplotlib.pyplot as plt
    >>> n = mt.zeros((200,200), dtype=complex)
    >>> n[60:80, 20:40] = mt.exp(1j*mt.random.uniform(0, 2*mt.pi, (20, 20)))
    >>> im = mt.fft.ifftn(n).real
    >>> plt.imshow(im.execute())
    <matplotlib.image.AxesImage object at 0x...>
    >>> plt.show()

    """
    a = astensor(a)
    axes = validate_fftn(a, s=s, axes=axes, norm=norm)
    op = TensorIFFTN(shape=s, axes=axes, norm=norm, dtype=np.dtype(np.complex_))
    return op(a)
